#lang "qb"

#cmdline "-gen gcc -O 3 -lib" 

option explicit

#include "fbgfx.bi"

#if __FB_LANG__="qb"
  #define i8 __byte
  #define u8 __ubyte
  #define i16 __short
  #define u16 __ushort
  #define i32 __long
  #define u32 __ulong
  #define pu8 u8 __ptr
  #define with __with
  #define screenres __screenres
  #define screenptr __screenptr
  #define screenlock __screenlock
  #define screenunlock __screenunlock
  #define cuint __cuint
  #define cast __cast
  #define cptr __cptr
  #define ptr __ptr
  #define zstring __zstring
  #define asm __asm  
  #define sizeof __sizeof
  #define allocate __allocate
  #define callocate __callocate
  #define deallocate __deallocate
  #define reallocate __reallocate
  #define dir __dir
  #define shl __shl
  #define shr __shr
  #define cvshort __cvshort
  #define setmouse __setmouse
  #define getmouse __getmouse
  #define multikey __multikey
  #define screensync __screensync
  #define flip __flip
  #define screencontrol __screencontrol
  #define imagecreate __imagecreate
  #define imagedestroy __imagedestroy
  #define iif __iif
#else      
  #include "crt.bi"
  #define u8 ubyte
  #define i16 short
  #define pu8 u8 ptr
  #define __sleep sleep
#endif

type pixel as u8

'$INCLUDE: 'DIRECTQB.BI'

declare sub puts cdecl alias "puts" (pz as zstring ptr)

#define LogFunc rem
'#define LogFunc dbgprint

sub dbgprint( s as string )
  dim as zstring ptr p : p = strptr(s)
  asm    
    push [p]
    call puts
    add esp, 4
  end asm
end sub

'----------------------------------------------------------------------------
' DIRECTQB.BI
'   Include file for the DirectQB library version 1.61
'   by Angelo Mottola - Enhanced Creations 1998-99
'
' This file contains constants and function declarations used by the library
' Always include this file into your own programs!
'
' Include file created on 11-10-1999 by the DirectQB Library Manager v1.2
' Run DQBMAN again to modify your DirectQB module settings
'
'  Read DIRECTQB.DOC for detailed informations on how to use the library
'----------------------------------------------------------------------------

'$DYNAMIC
defint A-Z

#macro Unimplemented()
  dbgprint "Unimplemented: '" & __FUNCTION__ & "'" : do: sleep: loop
  '!!! works generically but bugs freebasic? !!!
  '#if __FB_QUERY_SYMBOL__( 1 , __FUNCTION_NQ__ )=17
  '  __FUNCTION_NQ__ = ""
  '#elseif __FB_QUERY_SYMBOL__( 1 , __FUNCTION_NQ__ )
  '  __FUNCTION_NQ__ = 0
  '#endif
#endmacro  
#macro UnimplementedN()
  Unimplemented()
  __FUNCTION_NQ__ = 0
#endmacro  
#macro UnimplementedS()
  Unimplemented()
  __FUNCTION_NQ__ = ""  
#endmacro  

type DQB_MouseRange
  as i16 L,T,R,B
end type
type DQB_Context  
  bInitialized  as i8  
  iLayerCount   as u8
  iSoundCount   as u8
  bUpdatePal    as u8  
  iFPS          as i16
  iTexSize      as i16
  tMouseRange   as DQB_MouseRange  
  pLayer(15)    as image ptr  
  pFontText     as image ptr
  pXms(255)     as any ptr  
end type

static shared as DQB_Context tDQB

'Option Nogosub
#define _return(_N) __FUNCTION_NQ__ = (_N) : exit function

' Procedures from MAIN.OBJ:
function DQBinit (byval NumLayers, byval NumSounds, byval MemSize) 'OK
  LogFunc "DQBInit(NumLayers=" & NumLayers & " , NumSounds=" & NumSounds & " , MemSize=" & MemSize & ")"
  if tDQB.bInitialized then _return(4) 'Already Initialized
  tDQB.bInitialized = 1
  if NumLayers > 7 then NumLayers = 7
  tDQB.iLayerCount = NumLayers
  if NumSounds > 8 then NumSounds = 8
  tDQB.iSoundCount = NumSounds  
  tDQB.pLayer(0)=0
  dim as long i
  for i = 1 to NumLayers
    tDQB.pLayer(i) = callocate( 320*200l + sizeof(_OLD_HEADER) )        
    if tDQB.pLayer(i) = 0 then 
      DbgPrint("DBQInit: Failed to allocate layers")
      _return(3) 'failed to allocate layers
    end if
    with *tDqb.pLayer(i)                  
      .old.bpp = 8 : .old.width = 320 : .old.height = 200      
    end with
  next i
  _return(0) 'Success
end function
function DQBver% ()
  UnimplementedN()
end function
function DQBid$ ()
  UnimplementedS()  
end function
function DQBmapLayer% (byval Layer%)
  UnimplementedN()
end function  
sub DQBclose () 'OK
  dim as long i
  for i = 1 to 9
    if tDQB.pLayer(i) then deallocate( tDQB.pLayer(i) ) : tDQB.pLayer(i)=0
  next i
end sub
sub DQBpeek (byval DataSeg%, byval DataOff%, byval Offset&, byval Length%)
  Unimplemented()
end sub
sub DQBpoke (byval DataSeg%, byval DataOff%, byval Offset&, byval Length%)
  Unimplemented()
end sub
sub DQBsort (byval ArraySeg%, byval ArrayOff%, byval NumRecords%, byval RecordLen%, byval IndexOff%)
  Unimplemented()
end sub
function DQBangle% (byval x1%, byval y1%, byval x2%, byval y2%)
  UnimplementedN()
end function  
sub DQBinitVGA () 'OK
  LogFunc "DQBInitVGA()"
  __screencontrol(SET_GL_2D_MODE , OGL_2D_MANUAL_SYNC)
  __screencontrol(SET_GL_SCALE , 3)
  screenres 320,200,8,,GFX_OPENGL
end sub
sub DQBinitText () 'OK
  'Unimplemented()
  screen 0
end sub
function DQBsetBaseLayer (byval Layer) 'OK?
  LogFunc "DQBsetBaseLayer(Layer=" & Layer & ")"  
  _return(0) 'already allocated
end function
sub DQBcopyLayer (byval SourceLayer, byval DestLayer) 'OK
  LogFunc "DQBcopyLayer(SourceLayer=" & SourceLayer & " , DestLayer=" & DestLayer & ")"
  if SourceLayer > 7 then dbgprint("Source Layer: "+hex$(SourceLayer))
  if DestLayer > 7 then dbgprint("DestLayer: "+hex$(SourceLayer))
  dim as image ptr pSrc,pDst
  pDst=tDQB.pLayer(  DestLayer and 7)
  pSrc=tDQB.pLayer(SourceLayer and 7)
  if pSrc=pDst then dbgprint("DQBcopyLayer: Invalid copy Src=" & SourceLayer & " Dst=" & DestLayer) : exit sub
  if pSrc=0 then dbgprint("DQBcopyLayer: source is screen Src=" & SourceLayer & " Dst=" & DestLayer)
  put pDst,(0,0),pSrc,pset    
  if (DestLayer=0) then 
    'if DQBframeReady() then flip
    flip
  end if
end sub
sub DQBclearLayer (byval Layer) 'OK
  LogFunc "DQBclearLayer(Layer=" & Layer & ")"
  line tDQB.pLayer(Layer and 7),(0,0)-(319,199),0,bf
end sub
sub DQBwait (byval Times) 'OK
  LogFunc "DQBwait(Times=" & Times & ")"  
  dim as long N
  if tDQB.bUpdatePal = 1 then
    tDQB.bUpdatePal = 0
    'dim as double dTemp : dTemp = __timer 
    flip : Times -= 1
    'if abs(timer-dTemp) > (1/120) then Times -= 1
    'timer
  end if
  for N = 0 to Times-1
    screensync    
  next N  
end sub
sub DQBsetFrameRate (byval FPS) 'ok
  LogFunc "DQBsetFrameRate(FPS=" & FPS & ")"
  tDQB.iFPS = FPS  
end sub
function DQBframeReady () 'ok
  LogFunc "=DQBframeReady()"
  static as double dTMR
  if abs(timer-dTMR) > 1 then dTMR = timer : _return(0)    
  if abs(timer-dTMR) > (1/tDQB.iFPS) then dTMR += (1/tDQB.iFPS) : _return(-1)
  _return(0)
end function
function DQBerror$ ()
  UnimplementedS()
end function

' Procedures from DRAW.OBJ:
sub DQBcopyTransLayer (byval SourceLayer, byval DestLayer) 'OK
  LogFunc "DQBcopyTransLayer(SourceLayer=" & SourceLayer & " , DestLayer=" & DestLayer & ")"
  put tDQB.pLayer(DestLayer and 7),(0,0),tDQB.pLayer(SourceLayer and 7),trans
end sub
sub DQBcopyHitLayer (byval SourceLayer%, byval DestLayer%, byval Col%)
  Unimplemented()
end sub
sub DQBpset (byval Layer, byval x, byval y, byval Col) 'OK
  pset tDQB.pLayer(Layer and 7) , ( x,y ), col  
end sub
function DQBpoint(byval Layer, byval x, byval y) 'OK
  _return(point(x,y,tDQB.pLayer(Layer and 7)))
end function
sub DQBline (byval Layer%, byval x1%, byval y1%, byval x2%, byval y2%, byval Col%)
  Unimplemented()
end sub
sub DQBgline (byval Layer%, byval x1%, byval y1%, byval x2%, byval y2%, byval Col1%, byval Col2%)
  Unimplemented()
end sub
sub DQBellipse (byval Layer%, byval x%, byval y%, byval rx%, byval ry%, byval Col%)
  Unimplemented()
end sub
sub DQBbox (byval Layer%, byval x1%, byval y1%, byval x2%, byval y2%, byval Col%)
  Unimplemented()
end sub
sub DQBboxf (byval Layer%, byval x1%, byval y1%, byval x2%, byval y2%, byval Col%)
  Unimplemented()
end sub
sub DQBpaint (byval Layer%, byval x%, byval y%, byval Col%)
  Unimplemented()
end sub
sub DQBscroll (byval Layer%, byval dx%, byval dy%)
  Unimplemented()
end sub
sub DQBscrollArea (byval Layer%, byval x1%, byval y1%, byval x2%, byval y2%, byval Direction%)
  Unimplemented()
end sub
sub DQBsetTransPut ()
  Unimplemented()
end sub
sub DQBsetSolidPut ()
  Unimplemented()
end sub
sub DQBget (byval Layer, byval x1, byval y1, byval x2, byval y2, byval DummyBufSeg, byval pBuf as any ptr) 'OK
  logfunc "DQBget(Layer=" & Layer & " , " & "(" & x1 & "," & y1 & ")-(" & x2 & "," & y2 & "), pBuf=0x" & hex$(pBuf,8) & ")"
  if x1 > x2 then swap x1,x2
  if y1 > y2 then swap y1,y2
  dim as image ptr pImg   : pImg = pBuf
  dim as image ptr pLayer : pLayer = tDQB.pLayer(Layer and 7)  
  with pImg->old
    .bpp = 0 : .width = (x2-x1)+1 : .height = (y2-y1)+1
    put pImg,(0,0),pLayer,(x1,y1)-(x2,y2),pset
  end with
end sub
sub DQBput (byval Layer, byval x, byval y, byval DummyBufSeg, byval pBuf as any ptr) 'OK
  logfunc "DQBput(Layer=" & Layer & " , " & "(" & x & "," & y & "), pBuf=0x" & hex$(pBuf,8) & ")"
  put tDQB.pLayer(Layer and 7),(x,y),pBuf,pset
end sub

' Procedures from IMAGE.OBJ:
function DQBloadImage (byval Layer, byval x, byval y, FileName as string, Pal as string, byref imgWidth as integer, byref imgHeight as integer) 'OK
  
  LogFunc "DQBloadImage(Layer=" & Layer & " , x=" & x & " , y=" & y & " , FileName=" & FileName & ")"
  
  type PcxHeader field = 1
    Magic        as u8  '10
    Version      as u8  '3,4,5?
    Encode       as u8  '0=raw , 1=rle
    Bpp          as u8  '1,2,4,(8)
    MinX         as i16 
    MinY         as i16 
    MaxX         as i16 
    MaxY         as i16 
    HorzDPI      as i16
    VertDPI      as i16
    EgaPal(47)   as u8  'EGA Pal
    Reserved0    as u8
    Planes       as u8  '(1) 3, 4
    Pitch        as i16
    PalMode      as i16 '1/2 ???
    HorzRes      as i16
    VertRes      as i16
    Reserved(53) as u8
  end type
  
  dim pScr as pu8 , bLock as u8 : pScr = cptr(pu8,tDQB.pLayer(Layer and 7))
  if pScr=0 then pScr = screenptr : bLock = 1 else pScr += sizeof(_OLD_HEADER)
  dim as integer f : f = freefile()
  if open(FileName for binary access read as #f) then dbgPrint("DQBloadImage failed to open '"+FileName+"'"): _return( 1 ) 'File couldnt be opened
  dim as u16 uMagic : get #f,,uMagic
  select case uMagic
  case &h030A,&h040A,&h050A 'PCX v3-5
    dim as PcxHeader tHdr : get #f,1,tHdr
    #define iLeft x 
    #define iTop y
    const iWidth=320,iHeight=200 'dqb layer size
    with tHdr  
      'if .Magic  <> 10 then dbgprint "Not a .pcx file" : sleep : system
      if .Bpp    <> 8  then dbgprint "PCX: Unsupported bpp: " & .Bpp : close #f : _return(3)
      if .Planes <> 1  then dbgprint "PCX: Unsupported planes: " & .Planes : close #f : _return(3)
      if .Pitch <> ((.MaxX-.MinX)+1) then dbgprint "PCX: Warning Pitch differ: " & .Pitch          
      dim as long iPos,uAmount,N,iSrcX,iSrcY,iX,iY
      dim as u8 bBuffer(4095+768),uTmp,uCor
      uAmount = -1 : iY = iTop
      ImgWidth = .Pitch : ImgHeight=((.MaxY-.MinY)+1)      
      if iY < cuint(iHeight) then iX = iLeft else iX = iWidth  
      if bLock then screenlock  
      do
        #macro WritePixel(_Pix)
          if cuint(iX) < iWidth then pScr[iX] = _Pix
          iX += 1 : iSrcX += 1
          if iSrcX = ImgWidth then 
            iSrcX = 0 : pScr += iWidth : iY += 1 : iSrcY += 1          
            if iSrcY = ImgHeight then exit do
            if iY < cuint(iHeight) then iX = iLeft else iX = iWidth  
          end if
        #endmacro
        if iPos >= uAmount then
          uAmount = (uAmount+1)-iPos 'remaining
          if uAmount = 1 then bBuffer(0)=bBuffer(iPos)
          get #f,,bBuffer(uAmount),4096-uAmount,iPos
          uAmount += (iPos-1) : iPos = 0
        end if    
        uTmp = bBuffer(iPos) : iPos += 1
        if uTmp >= 192 then      
          uCor = bBuffer(iPos) : iPos += 1      
          for N = 193 to uTmp
            WritePixel(uCor)
          next N
        else
          WritePixel( uTmp )      
        end if
      loop
      uTmp = bBuffer(iPos) : iPos += 1 : uAmount += 1
      if uTmp <> 12 then get #f,lof(f)-768,uTmp : uAmount=0 : iPos=0
      if uTmp = 12 then    
        get #f,,bBuffer(uAmount),768-(uAmount-iPos)    
        Pal = space$(768)        
        for N = 0 to 767
          Pal[N] = bBuffer(iPos+N) shr 2 '0 to 63
        next N        
      else
        DbgPrint("PCX: Warning: no palette found")
      end if      
      if bLock then screenunlock  
    end with
    close #f: _return(0)
  case cvshort("BM")        'BMP
    close #f : dbgPrint("TODO: implement BMP"): _return(2)
  case &h0006               'BSV
    close #f : dbgPrint("TODO: implement BSV"): _return(2)
  case else                 'Bad/Unknown file format
    close #f : dbgPrint("Unknown image format"):_return(3)
  end select  
end function
function DQBsaveImage% (Layer%, x1%, y1%, x2%, y2%, FileName$, Pal$, Format%)
  UnimplementedN()
end function
function DQBplayFLI% (FileName$, BufLayer%, KeyStop%, LoopFlag%)
  UnimplementedN()
end function
function DQBopenFLI% (FileName$, Frames%, Speed%)
  UnimplementedN()
end function
sub DQBplayFLIstep (byval Layer%)
  Unimplemented()
end sub
sub DQBcloseFLI ()
  Unimplemented()
end sub

' Procedures from SPRITE.OBJ: 
function DQBsize% (byval x1%, byval y1%, byval x2%, byval y2%)
  UnimplementedN()
end function
sub DQBsetClipBox (byval x1%, byval y1%, byval x2%, byval y2%)
  Unimplemented()
end sub
sub DQBfPut (byval Layer%, byval x%, byval y%, byval DummyBufSeg%, byval BufOff as any __ptr)
  Unimplemented()
end sub
sub DQBsPut (byval Layer%, byval x%, byval y%, byval DummyBufSeg%, byval BufOff as any __ptr, byval NewWidth%, byval NewHeight%)
  Unimplemented()
end sub
sub DQBrPut (byval Layer as integer, byval x as integer, byval y as integer, byval DummyBufSeg as integer, byval BufOff as any ptr, byval Angle as integer, byval Zoom as integer) 'needs to rotate/zoom
  const PI256 = (atn(1)*4)/128  
  dim as image ptr pBuf : pBuf = BufOff
  dim as long iWid,iHei : iWid = pBuf->Old.Width : iHei = pBuf->Old.Height  
  dim as single fWH,fHH,fScale : fWH = iWid/2 : fHH = iHei/2 : fScale = Zoom/100
  dim as long iWidScale,iHeiScale : iWidScale = iWid*fScale : iHeiScale = iHei*fScale  
  
  dim as single HX,HY : HX=cos(PI256*-Angle)/fScale      : HY=sin(PI256*-Angle)/fScale
  dim as single VX,VY : VX=cos(PI256*(-Angle+64))/fScale : VY=sin(PI256*(-Angle+64))/fScale
  dim as single FX,FY , fLeft,fTop,fRight,fBottom
  #macro ChkLimit()
    if FX < fLeft   then fLeft   = FX
    if FX > fRight  then fRight  = FX
    if FY < fTop    then fTop    = fY
    if FY > fBottom then fBottom = fY
  #endmacro
  FX = HX*(iWidScale) : FY = HY*(iHeiScale)
  ChkLimit()
  FX = VX*(iWidScale) : FY = VY*(iHeiScale)
  ChkLimit()
  FX = (VX+HX)*(iWidScale) : FY = (VY+HY)*(iHeiScale)
  ChkLimit()  
  FX = fWH-((fLeft+fRight+.5)/2) : FY = fHH-((fTop+fBottom+.5)/2)
  'select case Angle
  'case  64 to 127 : FY -= .5
  'case 128 to 191 : FX -= .5 : FY -= .33
  'case 192 to 255 : FX -= .33 : FY += .33
  'end select  
  dim as u8 ptr pRow 
  if (Layer and 7) then
    pRow=cptr(u8 ptr,tDQB.pLayer(Layer and 7)) : if pRow=0 then exit sub
    pRow += sizeof(_OLD_HEADER)+y*320+x
  else
    screenlock : pRow=screenptr+y*320+x
  end if
  
  dim as u8 ptr pSrc : pSrc=BufOff+sizeof(_OLD_HEADER)  
  dim as single PX,PY
  dim as long nX,nY,iX,iY
  dim as u8 bPix
  for nY = 0 to iHeiScale-1
    PX=FX : PY=FY    
    for nX = 0 to iWidScale-1
      iX = int(PX) : iY = int(PY)
      if cuint(iX)<iWid andalso cuint(iY)<iHei then
        bPix = pSrc[iY*iWid+iX]
        if bPix then pRow[nX] = bPix
      end if
      PX += HX : PY += HY
    next nX
    FX += VX : FY += VY : pRow += 320
  next nY
  if (Layer and 7)=0 then screenunlock  
end sub
sub DQBxPut (byval SourceLayer%, byval x1%, byval y1%, byval x2%, byval y2%, byval DestLayer%, byval x%, byval y%)
  Unimplemented()
end sub
sub DQBmPut (byval Layer%, byval x%, byval y%, byval DummyBufSeg%, byval BufOff as any __ptr, byval flip%)
  Unimplemented()
end sub
sub DQBhPut (byval Layer%, byval x%, byval y%, byval DummyBufSeg%, byval BufOff as any __ptr, byval Col%)
  Unimplemented()
end sub
sub DQBtPut (byval Layer%, byval x%, byval y%, byval DummyBufSeg%, byval BufOff as any __ptr, byval BitMode%)
  Unimplemented()
end sub
sub DQBpPut (byval Layer%, byval x%, byval y%, byval DummyBufSeg%, byval BufOff as any __ptr, byval Pattern%)
  Unimplemented()
end sub
sub DQBputOver (byval DummyBackSeg%, byval BackOff as any __ptr, byval x%, byval y%, byval DummyBufSeg%, byval BufOff as any __ptr)
  Unimplemented()
end sub
sub DQBsetCollideMethod (byval Method%)
  Unimplemented()
end sub
function DQBcollide% (byval x1%, byval y1%, byval DummySeg1%, byval Off1 as any __ptr, byval x2%, byval y2%, byval DummySeg2%, byval Off2 as any __ptr)
  UnimplementedN()
end function
function DQBcollideOnLayer% (byval Layer%, byval x%, byval y%, byval DummySpriteSeg%, byval SpriteOff as any __ptr)
  UnimplementedN()
end function

' Procedures from BIT.OBJ:
function DQBsetBit% (byval Value%, byval bbit%)
  UnimplementedN()
end function
function DQBresetBit% (byval Value%, byval bbit%)
  UnimplementedN()
end function
function DQBreadBit% (byval Value%, byval bbit%)
  UnimplementedN()
end function
function DQBtoggleBit% (byval Value%, byval bbit%)
  UnimplementedN()
end function
function DQBshiftLeft% (byval Value%, byval nBits%)
  UnimplementedN()
end function
function DQBshiftRight% (byval Value%, byval nBits%)
  UnimplementedN()
end function

' Procedures from PALETTE.OBJ:
sub DQBsetCol (byval ColorIndex, byval Red, byval Green, byval Blue) 'OK
  palette ColorIndex, (Red*255)\63,(Green*255)\63,(Blue*255)\63  
end sub
sub DQBgetCol (byval ColorIndex, Red, Green, Blue) 'OK
  dim as long R,G,B
  palette get ColorIndex, R,G,B
  Red = R shr 2 : Green = G shr 2 : Blue = B shr 2
end sub
function DQBfindCol% (byval Red%, byval Green%, byval Blue%)
  UnimplementedN()
end function
function DQBfindPalCol% (Pal$, Red%, Green%, Blue%)
  UnimplementedN()
end function
sub DQBsetPal (Pal as string) 'OK
  LogFunc "DQBSetPal(Pal[])"
  dim as long N,I
  for N = 0 to (len(Pal)\3)-1
    palette N,(Pal[I+0] shl 0l) or (Pal[I+1] shl 8l) or (Pal[I+2] shl 16l) : I += 3
  next N  
  tDQB.bUpdatePal = 1
end sub
sub DQBgetPal (Pal$)
  Unimplemented()
end sub
sub DQBfadeIn (Pal as string)
  Unimplemented()
end sub
sub DQBfadeStepIn (Pal as string)  
  dim as long N,R,G,B,I,Red,Green,Blue
  for N=0 to 255
    palette get N,R,G,B : R shr= 2 : G shr= 2 : B shr= 2 
    Red = Pal[I] : Green = Pal[I+1] : Blue = Pal[I+2]: I += 3
    R += sgn(Red-R) : G += sgn(Green-G) : B += sgn(Blue-B)    
    palette N,(R*255)\63,(G*255)\63,(B*255)\63
  next N 
  flip
end sub
sub DQBfadeTo (byval Red%, byval Green%, byval Blue%)
  Unimplemented()
end sub
sub DQBfadeStepTo (byval Red as integer, byval Green as integer, byval Blue as integer) 'OK
  dim as long N,R,G,B
  for N=0 to 255
    palette get N,R,G,B : R shr= 2 : G shr= 2 : B shr= 2 
    R += sgn(Red-R) : G += sgn(Green-G) : B += sgn(Blue-B)    
    palette N,(R*255)\63,(G*255)\63,(B*255)\63
  next N   
  tDQB.bUpdatePal = 1
end sub
sub DQBpalOff ()
  Unimplemented()
end sub
sub DQBpalRotate (byval FirstCol%, byval LastCol%, byval RotDir%)
  Unimplemented()
end sub

enum TextStyle
  tsNONE       =  0 
  tsSOLID      =  1
  tsBOLD       =  2
  tsITALIC     =  4
  tsUNDERLINED =  8
  tsBLENDED    = 16
  tsTEXTURED   = 32 
end enum

' Procedures from FONT.OBJ:
sub DQBprint (Layer%, Text$, x%, y%, Col%)
  Unimplemented()
end sub
sub DQBprints (byval Layer as integer, Text as string, byval x as integer, byval y as integer, byval Col as integer, byval Style as integer) 'Need styles
  dim as image ptr pLayer : pLayer = tDQB.pLayer(Layer and 7)
  dim as long N,iBold,PX,iLen : iBold = (Style and BOLD)<>0 : iLen = len(Text)
  'if (Style and BLENDED) then DbgPrint "DQBprints: BLENDED style not implemented"
  'if (Style and ITALIC) then DbgPrint "DQBprints: ITALIC style not implemented"  
      
  if (Style and ITALIC) then 
    dim as image ptr pTemp : pTemp=ImageCreate(iLen*8,8)
    'if (Style and SOLID) andalso pLayer<>0 then line pTemp,(0,0)-step((not iBold)+(iLen*8),7),0,bf
    if (Style and TEXTURED) then
      Col = 0 : PX = 0 : for N = 0 to iLen-1 : put pTemp,(PX,0),tDQB.pFontText,xor : PX += 8 : next N
    end if  
    draw string pTemp,(0,0),Text,Col
    if iBold then draw string pTemp,(1,0),Text,Col
    if (Style and UNDERLINED) then line pTemp,(0,7)-step((not iBold)+(iLen*8),0),Col
    if (Style and TEXTURED) then
      PX = 0 : for N = 0 to iLen-1 : put pTemp,(PX,0),tDQB.pFontText,xor : PX += 8 : next N
    end if
    if (Style and SOLID) then
      put pLayer,(x-2,y  ),pTemp,(0,0)-(iLen*8-1,2),pset
      put pLayer,(x-1,y+2),pTemp,(0,3)-(iLen*8-1,4),pset
      put pLayer,(x  ,y+4),pTemp,(0,5)-(iLen*8-1,7),pset
    else
      put pLayer,(x-2,y  ),pTemp,(0,0)-(iLen*8-1,2),trans
      put pLayer,(x-1,y+2),pTemp,(0,3)-(iLen*8-1,4),trans
      put pLayer,(x  ,y+4),pTemp,(0,5)-(iLen*8-1,7),trans
    end if
    ImageDestroy(pTemp)
  else
    'if (Style and SOLID) andalso pLayer<>0 then line pLayer,(x,y)-step((not iBold)+(iLen*8),7),0,bf
    if (Style and TEXTURED) then
      Col = 0 : PX = x : for N = 0 to iLen-1 : put pLayer,(PX,y),tDQB.pFontText,xor : PX += 8 : next N
    end if  
    draw string pLayer,(x,y),Text,Col
    if iBold then draw string pLayer,(x+1,y),Text,Col
    if (Style and UNDERLINED) then line pLayer,(x,y+7)-step((not iBold)+(iLen*8),0),Col
    if (Style and TEXTURED) then
      PX = x : for N = 0 to iLen-1 : put pLayer,(PX,y),tDQB.pFontText,xor : PX += 8 : next N
    end if
  end if
  
end sub
function DQBlen% (Text$)
  UnimplementedN()
end function
sub DQBsetBIOSfont ()
  Unimplemented()
end sub
function DQBloadFont% (FileName$)
  UnimplementedN()
end function
sub DQBsetFont (Font$)
  Unimplemented()
end sub
sub DQBsetTextStyle (byval Style%)
  Unimplemented()
end sub
sub DQBsetTextBackCol (byval Col%)
  Unimplemented()
end sub
sub DQBsetTextSpacing (byval Spacing%)
  Unimplemented()
end sub
sub DQBsetTextBMap (byval BMap%)
  Unimplemented()
end sub
sub DQBsetFontTexture (byval DummyTextSeg, byval TextOff as any ptr) 'OK
  tDQB.pFontText = TextOff
end sub

' Procedures from DISK.OBJ:
function DQBdir$ (Mask$, Attributes%) 'OK
  _return( dir(Mask$,Attributes%) )
end function
function DQBdrive$ ()
  UnimplementedS()
end function
function DQBpath$ ()
  UnimplementedS()
end function
function DQBnumDrives% ()
  UnimplementedN()
end function
sub DQBsetDrive (NewDrive$)
  Unimplemented()
end sub
function DQBchDir (NewDir$)
  UnimplementedN()
end function

' Procedures from BLENDING.OBJ:
sub DQBfilterBox (byval Layer%, byval x1%, byval y1%, byval x2%, byval y2%, byval Col%, byval BMap%)
  Unimplemented()
end sub
sub DQBbPut (byval Layer%, byval x%, byval y%, byval DummyBufSeg%, byval BufOff as any __ptr, byval BMap%)
  Unimplemented()
end sub
sub DQBcopyBlendLayer (byval SourceLayer%, byval DestLayer%, byval BMap%)
  Unimplemented()
end sub
function DQBcreateBMap% (byval BMap%, byval FirstCol%, byval LastCol%)
  UnimplementedN()
end function
function DQBloadBMap% (BMap%, FileName$)
  UnimplementedN()
end function
function DQBsaveBMap% (BMap%, FileName$)
  UnimplementedN()
end function
sub DQBsetBMap (byval BMap%, byval ForeCol%, byval BackCol%, byval NewCol%)
  Unimplemented()
end sub
function DQBgetBMap% (byval BMap%, byval ForeCol%, byval BackCol%)
  UnimplementedN()
end function

' Procedures from KEYBOARD.OBJ:
sub DQBinstallKeyboard () 'OK
  rem nothing to do
end sub
sub DQBremoveKeyboard () 'OK
  rem nothing to do
end sub
function DQBkey (byval ScanCode)
  _return(multikey(ScanCode))
end function
function DQBreadKey% ()
  UnimplementedN()
end function
sub DQBwaitKey (byval ScanCode%)
  Unimplemented()
end sub
function DQBasc% (byval ScanCode%, byval ShiftFlag%)
  UnimplementedN()
end function
function DQBinkey$ ()
  UnimplementedS()
end function

' Procedures from JOYSTICK.OBJ:
function DQBjoyDetected% (byval JoyNum%)
  UnimplementedN()
end function
function DQBjoyX% (byval JoyNum%)
  UnimplementedN()
end function
function DQBjoyY% (byval JoyNum%)
  UnimplementedN()
end function
function DQBjoyMove% (byval JoyNum%, byval Direction%)
  UnimplementedN()
end function
function DQBjoyFire% (byval JoyNum%, byval Button%)
  UnimplementedN()
end function
sub DQBpollJoy (byval JoyNum%)
  Unimplemented()
end sub
sub DQBresetJoy ()
  Unimplemented()
end sub
sub DQBsetJoySens (byval Sens%)
  Unimplemented()
end sub

' Procedures from MOUSE.OBJ:
function DQBmouseDetected% () 'OK
  _return(0) 'Mouse Detected
end function
function DQBmouseX () 'OK
  dim as long X,Y
  getmouse X,Y
  with tDQB.tMouseRange 
    if X < .L then X = .L
    if X > .R then X = .R
  end with
  _return(X)
end function
function DQBmouseY () 'OK
  dim as long X,Y
  getmouse X,Y
  with tDQB.tMouseRange 
    if Y < .T then Y = .T
    if Y > .B then Y = .B
  end with  
  _return(Y)
end function
function DQBmouseLB () 'OK
  dim as long X,Y,B
  getmouse X,Y,,B
  DQBmouseLB = -(B and 1)  
end function
function DQBmouseRB () 'OK
  dim as long X,Y,B
  getmouse X,Y,,B
  DQBmouseRB = -((B and 1) shr 1)
end function
sub DQBsetMousePos (byval x, byval y) 'OK
  if x < tDQB.tMouseRange.L then x=tDQB.tMouseRange.L
  if y < tDQB.tMouseRange.T then y=tDQB.tMouseRange.T
  if x > tDQB.tMouseRange.R then x=tDQB.tMouseRange.R
  if y > tDQB.tMouseRange.B then y=tDQB.tMouseRange.B
  setmouse x,y  
end sub
sub DQBmouseShow ()
  Unimplemented()
end sub
sub DQBmouseHide ()
  Unimplemented()
end sub
sub DQBsetMouseRange (byval x1, byval y1, byval x2, byval y2) 'OK
  with tDQB.tMouseRange
    .L=x1 : .T=y1 : .R=x2 : .B=y2
  end with
end sub
sub DQBsetMouseShape (hotX%, hotY%, Shape$)
  Unimplemented()
end sub
sub DQBsetMouseSpeed (byval HorSpeed%, byval VerSpeed%)
  Unimplemented()
end sub
sub DQBresetMouse () 'OK
  with tDQB.tMouseRange 
    .L=0 : .T=0 : .R=319 : .B=199
  end with
  setmouse 160,100,1,0
end sub

' Procedures from SOUND.OBJ:
function DQBinstallSB% (byval VolActive%, byval Channels%, byval Freq%, byval BaseAddr%, byval IRQ%, byval DMA%)
  UnimplementedN()
end function
function DQBloadSound% (Slot%, FileName$)
  UnimplementedN()
end function
function DQBloadRawSound% (Slot%, FileName$, Offset&, Length&)
  UnimplementedN()
end function
function DQBinUse (byval Voice) 'dummy
  _return(0)
end function
sub DQBplaySound (byval SoundNum%, byval Voice%, byval Freq%, byval LoopFlag%)
  Unimplemented()
end sub
sub DQBsetVoiceVol (byval Voice%, byval NewVol%)
  Unimplemented()
end sub
sub DQBpauseSound ()
  Unimplemented()
end sub
sub DQBresumeSound ()
  Unimplemented()
end sub
sub DQBstopVoice (byval Voice%)
  Unimplemented()
end sub
sub DQBremoveSB ()
  Unimplemented()
end sub
sub DQBsetVolume (byval Volume%)
  Unimplemented()
end sub

' Procedures from 3D.OBJ:
sub DQBtri (byval Layer, byval x1, byval y1, byval x2, byval y2, byval x3, byval y3, byval Col)
  Unimplemented()  
end sub
sub DQBbtri (byval Layer%, byval x1%, byval y1%, byval x2%, byval y2%, byval x3%, byval y3%, byval Col%, byval BMap%)
  Unimplemented()
end sub
sub DQBgtri (byval Layer%, byval x1%, byval y1%, byval c1%, byval x2%, byval y2%, byval c2%, byval x3%, byval y3%, byval c3%)
  Unimplemented()
end sub
sub DQBbgtri (byval Layer%, byval x1%, byval y1%, byval c1%, byval x2%, byval y2%, byval c2%, byval x3%, byval y3%, byval c3%, byval BMap%)
  Unimplemented()
end sub
sub DQBttri (byval Layer, byval x1, byval y1, byval x2, byval y2, byval x3, byval y3, byval u1, byval v1, byval u2, byval v2, byval u3, byval v3, byval DummyTextureSeg, byval TextureOff as any ptr) 'OK
  
  dim as any ptr pLayer : pLayer = tDQB.pLayer(Layer and 7)
  dim as long TexSize : TexSize = tDQB.iTexSize  
  
  'line pLayer,(x1,y1)-(x2,y2),63
  'line pLayer,-(x3,y3),63
  'line PLayer,-(x1,y1),63
  'exit sub
  
  #define _Min( _A , _B ) iif( (_A) < (_B) , (_A) , (_B) )
  #define _Max( _A , _B ) iif( (_A) > (_B) , (_A) , (_B) )

  ' Bounding box for triangle
  dim as long minX : minX = (_Min(x1, _Min(x2, x3)))
  dim as long minY : minY = (_Min(y1, _Min(y2, y3)))
  dim as long maxX : maxX = (_Max(x1, _Max(x2, x3)))
  dim as long maxY : maxY = (_Max(y1, _Max(y2, y3)))
  
  dim as double InvSize : InvSize = 1/TexSize
  dim as double fu1,fu2,fu3 : fu1=u1*InvSize : fu2=u2*InvSize : fu3=u3*InvSize
  dim as double fv1,fv2,fv3 : fv1=v1*InvSize : fv2=v2*InvSize : fv3=v3*InvSize
  
  dim as double denom : denom = 1/(((y2-clng(y3)) * (x1-x3) + (x3-clng(x2)) * (y1-y3)))
  #define SampleTexture( _u , _v , _pTex ) _pTex[ ((cint((_v)*TexSize) and (TexSize-1))*texSize) + (cint((_u)*TexSize) and (TexSize-1)) ]
  
  if minX < 0 then minX = 0
  if minY < 0 then minY = 0
  if MaxX > 319 then MaxX = 319
  if MaxY > 199 then MaxY = 199
  
  dim as double fuStart, fvStart, fuStep , fvStep 
  fuStep  = (y2-y3)*denom    : fvStep  = (y3-y1)*denom
  fuStart = (minX-x3)*fuStep : fvStart = (minX-x3)*fvStep
  
  dim as pixel ptr pTex,pTgt : pTex = TextureOff+4    
  dim as pixel uPix  
  if pLayer=0 then screenlock : pTgt = screenptr else pTgt = pLayer+4
  pTgt += 320l*minY
  
  dim as double fu,fv , texU,texV
  dim as long x,y
  for y = minY to maxY
    fu = fuStart + ((x3-clng(x2))*(y-y3))*denom
    fv = fvStart + ((x1-clng(x3))*(y-y3))*denom
    for x = minX to maxX          
      ' Check if the pixel is inside the triangle
      if fu >= 0 Andalso fv >= 0 then
        dim as single fw : fw = 1.0 - fu - fv
        if fw >= 0 then
          ' Interpolate UV coordinates
          texU = fu*fu1 + fv*fu2 + fw*fu3
          texV = fu*fv1 + fv*fv2 + fw*fv3                
          uPix = SampleTexture( texU,texV , pTex )
          if uPix then pTgt[x] = uPix
        end if
      end if
      ' update barycentric coordinates
      fu += fuStep : fv += fvStep
    next x
    pTgt += 320
  next y  
  
  'line pLayer, (x1,y1)-(x2,y2), 63 'point(u1 and (TexSize-1),v1 and (TexSize-1),pTex)
  'line pLayer, -(x3,y3), 63 'point(u2 and (TexSize-1),v2 and (TexSize-1),pTex)
  'line pLayer, -(x1,y1), 63 'point(u3 and (TexSize-1),v3 and (TexSize-1),pTex)
  
  if pLayer=0 then screenlock
  
end sub
sub DQBbttri (byval Layer%, byval x1%, byval y1%, byval x2%, byval y2%, byval x3%, byval y3%, byval u1%, byval v1%, byval u2%, byval v2%, byval u3%, byval v3%, byval DummyTextureSeg%, byval TextureOff as any __ptr, byval BMap%)
  Unimplemented()
end sub
sub DQBfttri (byval Layer%, byval x1%, byval y1%, byval x2%, byval y2%, byval x3%, byval y3%, byval u1%, byval v1%, byval u2%, byval v2%, byval u3%, byval v3%, byval DummyTextureSeg%, byval TextureOff as any __ptr)
  Unimplemented()
end sub
sub DQBsetTextureSize (byval Size) 'OK
  dim as long iSz : iSz = 1
  while (iSz+iSz) <= Size
    iSz += iSz
  wend
  tDQB.iTexSize = iSz      
  if iSz <> Size then
    DbgPrint("Texture size=" & iSz & " (requested " & Size & ")")
  end if
end sub
sub xmsapi () 'OK (nothing)
  rem :)
end sub
sub XmsFree (byref fmem as long, byref tmem as long) 'OK
  LogFunc "XmsFree(&fmem , &tmem)"
  fmem = fre(-1) : tmem = fre(-1)
end sub
function movefromxms ( byval Kb as long , byval hHandle , byval xmsoffs as long , byval DummySegm , byval Offs as any ptr ) 'OK
  LogFunc "movefromxms(Kb=" & Kb & " , hHandle=" & hHandle & " , xmsoffs=" & xmsoffs & " , Offs=0x" & hex$(Offs,8) & ")"
  if cuint(hHandle)>255 orelse tDQB.pXms(hHandle)=0 then DbgPrint("movefromxms: bad handle: " & hhandle) : _return(1)  
  fb_memcopy( *cptr(u8 ptr,Offs) , *cptr(u8 ptr,tDQB.pXms(hHandle)+xmsoffs) , Kb )
  _return(0) 
end function
function move2xms ( byval Kb as long , byval DummySegm as integer , byval Offs as any ptr , byval hHandle as integer , byval xmsoffs as long ) 'OK
  LogFunc "move2xms(Kb=" & Kb & " , Offs=0x" & hex$(Offs,8) & " , hHandle=" & hHandle & " , xmsoffs=" & xmsoffs & ")"
  if cuint(hHandle)>255 orelse tDQB.pXms(hHandle)=0 then DbgPrint("movefromxms: bad handle: " & hhandle) : _return(1)
  fb_memcopy( *cptr(u8 ptr,tDQB.pXms(hHandle)+xmsoffs) , *cptr(u8 ptr,Offs) , Kb )
  _return(0)  
end function
function XmsdeAlloc (byval hHandle as integer) 'OK
  LogFunc "XmsdeAlloc( hHandle=" & hHandle & ")"
  if cuint(hHandle)>255 then _return(0)
  if tDQB.pXms(hHandle) = 0 then _return(0)
  deallocate( tDQB.pXms(hHandle) )
  tDQB.pXms(hHandle) = 0
  _return( hHandle )
end function
function XmsAlloc (byval Kb as integer) 'OK
  dim as long i
  for i = 1 to 255
    if tDQB.pXms(i)=0 then
      tDQB.pXms(i) = callocate( Kb*1024l ) 
      if tDQB.pXms(i) then 
        LogFunc i & "=XmsAlloc( Kb=" & kB & ")"
        _return( i )
      end if
      DbgPrint "(failed) XmsAlloc( Kb=" & kB & ")"
      _return(0)      
    end if
  next i
  DbgPrint "(failed) XmsAlloc( Kb=" & kB & ")"
  _return( 0 )
end function
function XmsVersion ()
  UnimplementedN()
end function
function XmsRealloc (byval hHandle as integer, byval Kb as integer) 'OK
  LogFunc "XmsRealloc( hHandle=" & hHandle & " , Kb=" & kB & ")"
  if cuint(hHandle)>255 then _return(0)
  if tDQB.pXms(hHandle) = 0 then _return(0)
  dim pNew as any ptr
  pNew = reallocate( tDQB.pXms(hHandle%) , Kb*1024 )
  if pNew then    
    tDQB.pXms(hHandle) = pNew
    _return(hHandle)
  end if
  _return(0)    
end function
